from .._base import Endpoint
from ...settings import json_lib as json


class AnswerInlineQuery(Endpoint):
    def __init__(self, inline_query_id, results):
        super().__init__()
        self.args['inline_query_id'] = inline_query_id
        self.args['results'] = json.dumps([x.args for x in results])

    def cache_time(self, cache_time):
        self.args['cache_time'] = cache_time
        return self

    def is_personal(self):
        self.args['is_personal'] = True
        return self

    def next_offset(self, next_offset):
        self.args['next_offset'] = next_offset
        return self

    def switch_pm_text(self, switch_pm_text):
        self.args['switch_pm_text'] = switch_pm_text
        return self

    def switch_pm_parameter(self, switch_pm_parameter):
        self.args['switch_pm_parameter'] = switch_pm_parameter
        return self



def _get_length(data_dict):
    track = data_dict.get('Video')
    if track is None:
        return

    width = track.get('Width')
    height = track.get('Height')
    if width < height:
        return width
    else:
        return height


attributes = {
    'audio': [
        ('duration', 'General', 'Duration'),
        ('performer', 'General', 'Performer'),
        ('title', 'General', 'Track')
    ],
    'video': [
        ('duration', 'General', 'Duration'),
        ('width', 'Video', 'Width'),
        ('height', 'Video', 'Height')
    ],
    'animation': [
        ('duration', 'General', 'Duration'),
        ('width', 'Video', 'Width'),
        ('height', 'Video', 'Height')
    ],
    'voice': [
        ('duration', 'General', 'Duration')
    ],
    'video_note': [
        ('duration', 'General', 'Duration'),
        ('length', _get_length, None),
    ]
}



from ._base import Endpoint
from ._mixins import DisableNotification


class ForwardMessage(Endpoint, DisableNotification):
    def __init__(self, chat_id, from_chat_id, message_id):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['from_chat_id'] = from_chat_id
        self.args['message_id'] = message_id



from abc import ABC, abstractmethod

from .bot import Bot


class Trigger(ABC):
    __slots__ = []

    @classmethod
    @abstractmethod
    async def match(cls, update, bot: Bot):
        """Returns a truthy value if update matches, a falsy value if it doesn't. Async supported."""
        pass

    @classmethod
    @abstractmethod
    async def handle(cls, update, bot: Bot):
        pass



# flake8: noqa

from .endpoints import AnswerInlineQuery
from .input_message_content import Text as InputTextMessageContent, \
    Location as InputLocationMessageContent, \
    Venue as InputVenueMessageContent, \
    Contact as InputContactMessageContent
from .query_results import Article as InlineQueryResultArticle, \
    Photo as InlineQueryResultPhoto, \
    CachedPhoto as InlineQueryResultCachedPhoto, \
    Gif as InlineQueryResultGif, \
    CachedGif as InlineQueryResultCachedGif, \
    Mpeg4Gif as InlineQueryResultMpeg4Gif, \
    CachedMpeg4Gif as InlineQueryResultCachedMpeg4Gif, \
    Video as InlineQueryResultVideo, \
    CachedVideo as InlineQueryResultCachedVideo, \
    Audio as InlineQueryResultAudio, \
    CachedAudio as InlineQueryResultCachedAudio, \
    Voice as InlineQueryResultVoice, \
    CachedVoice as InlineQueryResultCachedVoice, \
    Document as InlineQueryResultDocument, \
    CachedDocument as InlineQueryResultCachedDocument, \
    Location as InlineQueryResultLocation, \
    Venue as InlineQueryResultVenue, \
    Contact as InlineQueryResultContact, \
    Game as InlineQueryResultGame, \
    CachedSticker as InlineQueryResultCachedSticker



from abc import ABC

try:
    from libmediainfo_cffi import MediaInfo
except Exception:
    pass

from ._attributes import attributes
from ..._utils.casing import camel_to_snake
from ..._utils.multipart_encoder import MultipartEncoder
from ...settings import json_lib as json


class FileMixin(ABC):
    __slots__ = []
    file_type: str
    attributes: dict

    args: dict
    value: str or list
    is_data: bool
    use_cache: bool


class File(ABC):
    __slots__ = ['file_type', 'attributes', 'args', 'value', 'is_data', 'use_cache']
    cache = {}

    def __init__(self, file, is_path=True):
        """Generic File class, supports automatic metadata reading via MediaInfo bindings and caching

        :param file: a path (default) or a valid string recognizable by the Telegram endpoint (an url or a file id)
        :param is_path: whether file is a path (default) or anything else
        """

        # cache is a dict {file_path: file_id} and should be updated with the trigger callback
        classname = self.__class__.__name__
        self.file_type = camel_to_snake(classname)
        self.args = {}
        self.file = file
        self.attributes = {}
        self.use_cache = False

        # if we received a path we first check we have the file inside the cache dict
        # if we do then we treat this as a File with "is_path" set to false and we use the file_id we've found as
        # else we store encode the file as multipart/form-data
        if is_path is True:
            file_id = File.cache.get(self.file)
            if file_id is not None:
                self.is_data = False
                self.value = file_id
            else:
                self.is_data = True
                self.value = [MultipartEncoder.encode_file(self.file_type, self.file)]
        else:
            self.is_data = False
            self.value = self.file

    def read_metadata(self):
        """Read metadata from the file (if it's a file)"""

        if not self.is_data or self.attributes:
            return self

        data = MediaInfo.read_metadata(self.file, Inform='JSON')

        data = json.loads(data)
        data_dict = {}

        # redrder the data to avoid O(tracks number) behavior while checking for tracks of a given type
        # also we use only the first track found, you can override this method if this doesn't work for you
        for track in data['media']['track']:
            _type = track['@type']

            if data_dict.get(_type) is None:
                data_dict[_type] = track

        for attribute, _type, key in attributes[self.file_type]:
            if callable(_type):
                attribute_value = _type(data)
            else:
                track = data_dict.get(_type)
                if track is None:
                    continue

                attribute_value = track.get(key)

            if attribute_value is None:
                continue

            self.attributes[attribute] = attribute_value

        return self

    def with_cache(self):
        """Check if a file_id can be used instead of a multipart/form-data request"""

        if not self.is_data:
            return self

        if not self.use_cache:
            self.use_cache = True
            return self

        file_id = File.cache.get(self.file)

        if file_id is not None:
            self.is_data = False
            self.value = file_id
            self.attributes = {}

        return self

    @staticmethod
    def _get_file_id(response):
        """If the telegram response is (for example) an array with different resolutions, here you can return the
        correct entry to the update_cache function
        """

        return response

    def update_cache(self, response, index=None):
        """Updates the object cache with the file ids"""

        # the index param is used for media groups
        # add correct file type automatically or ask the user to pass the type using bot.find_update_type
        file_id = File.cache.get(self.file)
        if file_id is None:
            response = response['result']
            if index is not None:
                response = response[index][self.file_type]
            else:
                response = response[self.file_type]

            File.cache[self.file] = self._get_file_id(response)['file_id']



from ._base import BaseMixin, Endpoint
from ._mixins import DisableNotification, ReplyMarkup, ReplyToMessageId, chat_or_inline_message


class LocationMixin(BaseMixin):
    def live_period(self, live_period):
        assert 60 <= live_period <= 86400

        self.args['live_period'] = live_period
        return self


class SendLocation(Endpoint, DisableNotification, LocationMixin, ReplyMarkup, ReplyToMessageId):
    def __init__(self, chat_id, latitude, longitude):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['latitude'] = latitude
        self.args['longitude'] = longitude


class EditMessageLiveLocation(Endpoint, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id, latitude, longitude):
        super().__init__()
        self.args['latitude'] = latitude
        self.args['longitude'] = longitude
        self.args.update(chat_or_inline_message(chat_or_inline_message_id, message_id))


class StopMessageLiveLocation(Endpoint, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id):
        super().__init__()
        self.args.update(chat_or_inline_message(chat_or_inline_message_id, message_id))



from .._base import Endpoint
from .._mixins import Caption, DisableNotification, ParseMode, ReplyToMessageId, chat_or_inline_message
from ..reply_markup import ReplyMarkup
from ..._utils.multipart_encoder import MultipartEncoder
from ...settings import json_lib as json


class SendMedia(Endpoint, DisableNotification, ReplyMarkup, ReplyToMessageId):
    def __init__(self, file, chat_id):
        super().__init__()
        self.endpoint = f'send{file.__class__.__name__}'
        self.args['chat_id'] = chat_id
        self.args.update(file.args)
        self.args.update(file.attributes)

        if file.is_data:
            self.method = 'POST'
            multipart_encoder = MultipartEncoder()
            multipart_encoder.files = file.value
            boundary, self.data = multipart_encoder.encode()
            self.headers['content-type'] = f'multipart/form-data; boundary={boundary}'
        else:
            self.args[file.file_type] = file.value


class SendMediaGroup(Endpoint, DisableNotification, ReplyToMessageId):
    media_types = ['photo', 'video']

    def __init__(self, chat_id, files):
        super().__init__()

        self.__multipart_encoder = None
        self.method = 'GET'
        media = []

        for file in files:
            assert file.file_type in self.media_types

            serialized, files = file.to_input_media()
            media.append(serialized)
            if files is not None:
                self.method = 'POST'
                self.multipart_encoder.files.extend(files)

        self.args['chat_id'] = chat_id
        self.args['media'] = json.dumps(media)

        if self.method == 'POST':
            boundary, self.data = self.multipart_encoder.encode()
            self.headers['content-type'] = f'multipart/form-data; boundary={boundary}'

    @property
    def multipart_encoder(self):
        if self.__multipart_encoder is None:
            self.__multipart_encoder = MultipartEncoder()
        return self.__multipart_encoder


class GetFile(Endpoint):
    # https://api.telegram.org/file/bot<token>/<file_path>
    def __init__(self, file_id):
        super().__init__()
        self.args['file_id'] = file_id


class EditMessageCaption(Endpoint, Caption, ParseMode, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id):
        super().__init__()
        self.args.update(chat_or_inline_message(chat_or_inline_message_id, message_id))


class EditMessageMedia(Endpoint, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id, media):
        super().__init__()
        self.args['media'] = media.to_input_media()
        self.args.update(chat_or_inline_message(chat_or_inline_message_id, message_id))



from ._base import Endpoint


class DeleteMessage(Endpoint):
    def __init__(self, chat_id, message_id):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['message_id'] = message_id



import asyncio
from typing import NamedTuple
from urllib.parse import urlencode

from ._utils.case_insensitive_dict import CaseInsensitiveDict
from ._utils.connection_pool import HttpConnectionPool


class BotRequest(NamedTuple):
    endpoint: str
    method: str = 'GET'
    params: dict = {}
    data: bytes = None
    headers: CaseInsensitiveDict = CaseInsensitiveDict()


class Bot:
    update_types = [
        ('message', 'message'),
        ('edited_message', 'message'),
        ('channel_post', 'message'),
        ('edited_channel_post', 'message'),
        ('inline_query', 'inline_query'),
        ('chosen_inline_result', 'chosen_inline_result'),
        ('callback_query', 'callback_query'),
        ('shipping_query', 'shipping_query'),
        ('pre_checkout_query', 'pre_checkout_query')
    ]

    # optional features
    check_update = None

    def __init__(self, token, *, loop=None, connection_pool_size=5):
        """
        The Class(TM).

        :param token: The Telegram-given token
        :param loop: The loop the bot is run into (if it's not asyncio.get_event_loop())
        """

        self.token = token

        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        self.base_path = f'bot{token}/'
        self.connection_pool = HttpConnectionPool('api.telegram.org', loop=loop, size=connection_pool_size)
        self.triggers = []
        self.update_queue = asyncio.PriorityQueue(loop=self.loop)

    async def api_request(self, request: BotRequest):
        """
        Makes a Telegram request with the params given in the BotRequest

        :param request: A BotRequest
        :return: The response from the server
        """

        endpoint = request.endpoint
        params = request.params
        data = request.data
        headers = request.headers

        if params:
            path = f'{self.base_path}{endpoint}?{urlencode(params)}'
        else:
            path = f'{self.base_path}{endpoint}'

        return await self.connection_pool.request(path=path, method=request.method, body=data, headers=headers)

    def get_type_and_flavor(self, update):
        for _type, flavor in self.update_types:
            if _type in update:
                update['_type'] = _type
                update['_flavor'] = flavor
                return
        else:
            update['_type'] = None
            update['_flavor'] = None

    async def start(self):
        """Main loop"""

        while True:
            _, update = await self.update_queue.get()
            self.loop.create_task(self.handle_update(update))

    async def handle_update(self, update):
        # check the update if the function is implemented and skip if it's not passed
        # note that while the function is called check_update it's basically a pre-processing hook, so if you have to
        # call get_type_and_flavor do it here
        check_update = self.check_update
        if check_update:
            is_check_passed = await check_update(update)
            if is_check_passed is not True:
                return

        for trigger in self.triggers:
            is_update_matched = await trigger.match(update, self)
            if is_update_matched:
                return await trigger.handle(update, self)

    def push_update(self, update):
        """
        Pushes an update (already json decoded) into the queue.

        :param update: The update to be pushed in the queue
        """

        self.update_queue.put_nowait((update['update_id'], update))

    def trigger(self, trigger):
        """
        Decorates a Trigger, inserting it into the bot check list
        """

        # maybe trigger may already be instances? Or we can call the methods directly?
        # or use factory methods
        self.triggers.append(trigger)
        return trigger



from ._base import Endpoint
from ._mixins import ReplyMarkup, chat_or_inline_message


class InlineKeyboard:
    __slots__ = ['current_row', 'rows']
    button_types = {
        'url', 'callback_data', 'switch_inline_query', 'switch_inline_query_current_chat', 'callback_game',
        'pay'
    }
    must_be_first_button_types = {'callback_game', 'pay'}

    def __init__(self):
        self.current_row = None
        self.rows = []

    def serialize(self):
        return {'inline_keyboard': self.rows}

    def add_row(self):
        row = []
        self.rows.append(row)
        self.current_row = row
        return self

    def add_button(self, text, button_type, button_value):
        assert button_type in InlineKeyboard.button_types

        if button_type in InlineKeyboard.must_be_first_button_types:
            assert self.current_row

        self.current_row.append({
            'text': text,
            button_type: button_value
        })
        return self


class ReplyKeyboard:
    __slots__ = ['buttons', 'options']

    def __init__(self):
        self.buttons = []
        self.options = {}

    def serialize(self):
        return {
            'keyboard': self.buttons,
            **self.options
        }

    def resize_keyboard(self):
        self.options['resize_keyboard'] = True
        return self

    def one_time_keyboard(self):
        self.options['one_time_keyboard'] = True
        return self

    def selective(self):
        self.options['selective'] = True
        return self

    def add_button(self, text, request_contact=False, request_location=False):
        button = {
            'text': text
        }

        if request_contact:
            button['request_contact'] = True

        if request_location:
            button['request_location'] = True

        self.buttons.append(button)
        return self


class ReplyKeyboardRemove:
    __slots__ = ['options']

    def __init__(self):
        self.options = {}

    def serialize(self):
        return {
            'remove_keyboard': True,
            **self.options
        }

    def selective(self):
        self.options['selective'] = True
        return self


class ForceReply:
    __slots__ = ['options']

    def __init__(self):
        self.options = {}

    def serialize(self):
        return {
            'force_reply': True,
            **self.options
        }

    def selective(self):
        self.options['selective'] = True
        return self


class AnswerCallbackQuery(Endpoint):
    def __init__(self, callback_query_id):
        super().__init__()
        self.args['callback_query_id'] = callback_query_id

    def text(self, text):
        self.args['text'] = text
        return self

    def show_alert(self):
        self.args['show_alert'] = True
        return self

    def url(self, url):
        self.args['url'] = url
        return self

    def cache_time(self, cache_time):
        self.args['cache_time'] = cache_time
        return self


class EditMessageReplyMarkup(Endpoint, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id):
        super().__init__()
        self.args.update(args=chat_or_inline_message(chat_or_inline_message_id, message_id))



from ._base import BaseMixin, Endpoint
from ._mixins import DisableNotification, ReplyMarkup, ReplyToMessageId


class ContactMixin(BaseMixin):
    def last_name(self, last_name):
        self.args['last_name'] = last_name
        return self

    def vcard(self, vcard):
        self.args['vcard'] = vcard
        return self


class SendContact(Endpoint, ContactMixin, DisableNotification, ReplyMarkup, ReplyToMessageId):
    def __init__(self, chat_id, phone_number, first_name):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['phone_number'] = phone_number
        self.args['first_name'] = first_name



def build_request(host, path, method='GET', headers=None, body=None):
    if headers is None:
        headers = {}

    headers['Host'] = host
    string = f'{method} /{path} HTTP/1.1\r\n'
    string += '\r\n'.join(f'{k}: {v}' for k, v in headers.items())
    string += '\r\n\r\n'

    if body is not None:
        string += body
        if 'content-length' not in headers:
            headers['content-length'] = len(body)

    return string.encode('utf-8')



from ._base import Endpoint


class SendChatAction(Endpoint):
    actions = {'typing', 'upload_photo', 'record_video', 'upload_video', 'record_audio', 'upload_audio',
               'upload_document', 'find_location', 'record_video_note', 'upload_video_note'}

    def __init__(self, chat_id, action):
        assert action in SendChatAction.actions

        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['action'] = action



from ._base import BaseMixin
from ..settings import json_lib as json


def chat_or_inline_message(chat_or_inline_message_id, message_id):
    if message_id is not None:
        return {
            'chat_id': chat_or_inline_message_id,
            'message_id': message_id
        }
    else:
        return {
            'inline_message_id': chat_or_inline_message_id
        }


class Caption(BaseMixin):
    def caption(self, caption):
        assert len(caption) < 1025

        self.args['caption'] = caption
        return self


class ReplyToMessageId(BaseMixin):
    def reply_to_message_id(self, message_id):
        self.args['reply_to_message_id'] = message_id
        return self


class DisableNotification(BaseMixin):
    def disable_notification(self):
        self.args['disable_notification'] = True
        return self


class ParseMode(BaseMixin):
    parse_modes = {'Markdown', 'HTML'}

    def parse_mode(self, parse_mode):
        assert parse_mode in ParseMode.parse_modes

        self.args['parse_mode'] = parse_mode
        return self


class ReplyMarkup(BaseMixin):
    def reply_markup(self, reply_markup):
        self.args['reply_markup'] = json.dumps(reply_markup.serialize())
        return self



from ._base import Endpoint


class SetWebhook(Endpoint):
    update_types = {
        'message', 'edited_message', 'channel_post', 'edited_channel_post', 'inline_query', 'chosen_inline_result',
        'callback_query', 'shipping_query', 'pre_checkout_query'
    }

    def __init__(self, url):
        super().__init__()
        self.args['url'] = url

    def certificate(self, certificate):
        self.args['certificate'] = certificate
        return self

    def max_connections(self, max_connections):
        self.args['max_connections'] = max_connections
        return self

    def allowed_updates(self, allowed_updates):
        for update_type in allowed_updates:
            assert update_type in SetWebhook.update_types

        self.args['allowed_updates'] = allowed_updates
        return self


class DeleteWebhook(Endpoint):
    pass


class GetWebhookInfo(Endpoint):
    pass



from ._base import BaseMixin, Endpoint
from ._mixins import DisableNotification, ParseMode, ReplyToMessageId, chat_or_inline_message
from .reply_markup import ReplyMarkup


class TextMixin(BaseMixin):
    def disable_web_page_overview(self):
        self.args['disable_web_page_overview'] = True
        return self


class SendMessage(Endpoint, DisableNotification, ParseMode, ReplyMarkup, ReplyToMessageId, TextMixin):
    def __init__(self, chat_id, text):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['text'] = text


class EditMessageText(Endpoint, ParseMode, ReplyMarkup, TextMixin):
    def __init__(self, chat_or_inline_message_id, message_id, text):
        super().__init__()
        self.args['text'] = text
        self.args.update(chat_or_inline_message(chat_or_inline_message_id, message_id))



from ._base import Endpoint
from ._mixins import DisableNotification
from .._utils.multipart_encoder import MultipartEncoder


class ExportChatInviteLink(Endpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args['chat_id'] = chat_id


class SetChatPhoto(Endpoint):
    def __init__(self, chat_id, photo):
        super().__init__()
        self.method = 'POST'
        self.args['chat_id'] = chat_id

        self.multipart_encoder = MultipartEncoder()
        self.multipart_encoder.add_file('photo', photo)
        boundary, self.data = self.multipart_encoder.encode()
        self.headers['content-type'] = f'multipart/form-data; boundary={boundary}'


class DeleteChatPhoto(Endpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args['chat_id'] = chat_id


class SetChatTitle(Endpoint):
    def __init__(self, chat_id, title):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['title'] = title


class SetChatDescription(Endpoint):
    def __init__(self, chat_id, description):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['description'] = description


class PinChatMessage(Endpoint, DisableNotification):
    def __init__(self, chat_id, message_id):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['message_id'] = message_id


class UnpinChatMessage(Endpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args['chat_id'] = chat_id


class LeaveChat(Endpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args['chat_id'] = chat_id


class GetChat(Endpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args['chat_id'] = chat_id


class GetChatAdministrators(Endpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args['chat_id'] = chat_id


class GetChatMemberCount(Endpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args['chat_id'] = chat_id


class GetChatMember(Endpoint):
    def __init__(self, chat_id, user_id):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['user_id'] = user_id


class SetChatStickerSet(Endpoint):
    def __init__(self, chat_id, sticker_set_name):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['sticker_set_name'] = sticker_set_name


class DeleteChatStickerSet(Endpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args['chat_id'] = chat_id



from abc import ABC

from .._utils.case_insensitive_dict import CaseInsensitiveDict
from ..bot import BotRequest


class BaseMixin(ABC):
    __slots__ = []
    args: dict


# unused currently
class EndpointMixin(ABC):
    method: str
    endpoint: str
    args: dict
    data: bytes or None
    headers: CaseInsensitiveDict


class Endpoint(ABC):
    __slots__ = ['method', 'endpoints', 'args', 'data', 'headers']

    def __init__(self):
        self.method = 'GET'
        classname = self.__class__.__name__
        self.endpoint = f'{classname[0].lower()}{classname[1:]}'

        self.args = {}
        self.data = None
        self.headers = CaseInsensitiveDict()

    def serialize(self):
        return BotRequest(
            method=self.method,
            endpoint=self.endpoint,
            params=self.args,
            data=self.data,
            headers=self.headers
        )



from ._base import InputMessageContent
from .._mixins import ParseMode
from ..contact import ContactMixin
from ..location import LocationMixin
from ..text import TextMixin
from ..venue import VenueMixin


class Text(InputMessageContent, ParseMode, TextMixin):
    def __init__(self, message_text):
        self.args = {
            'message_text': message_text
        }


class Location(InputMessageContent, LocationMixin):
    def __init__(self, latitude, longitude):
        self.args = {
            'latitude': latitude,
            'longitude': longitude
        }


class Venue(InputMessageContent, VenueMixin):
    def __init__(self, latitude, longitude, title, address):
        self.args = {
            'latitude': latitude,
            'longitude': longitude,
            'title': title,
            'address': address
        }


class Contact(InputMessageContent, ContactMixin):
    def __init__(self, phone_number, first_name):
        self.args = {
            'phone_number': phone_number,
            'first_name': first_name
        }



# flake8: noqa

from .chat import ExportChatInviteLink, SetChatPhoto, DeleteChatPhoto, SetChatTitle, SetChatDescription, PinChatMessage, UnpinChatMessage, LeaveChat, GetChat, GetChatAdministrators, GetChatMemberCount, GetChatMember, SetChatStickerSet, DeleteChatStickerSet
from .chat_action import SendChatAction
from .chat_member import KickChatMember, UnbanChatMember, RestrictChatMember, PromoteChatMember
from .contact import SendContact
from .delete import DeleteMessage
from .file import *
from .forward import ForwardMessage
from .inline import *
from .location import SendLocation, EditMessageLiveLocation, StopMessageLiveLocation
from .reply_markup import InlineKeyboard, ReplyKeyboard, ReplyKeyboardRemove, ForceReply, AnswerCallbackQuery, EditMessageReplyMarkup
from .text import SendMessage, EditMessageText
from .user import GetMe, GetUserProfilePhotos
from .venue import SendVenue
from .webhook import SetWebhook, DeleteWebhook, GetWebhookInfo


# DO NOT STAR IMPORT IF IT'S NOT FROM PACKAGES OR YOU'LL IMPORT LITERALLY EVERYTHING



# these functions do not give correct output in edge cases, if those need to be handled they will be updated
# accordingly


def camel_to_snake(string):
    parts = []
    word = []

    for char in string:
        if char.isupper():
            _word = ''.join(word)
            parts.append(_word)
            word = [char.lower()]
        else:
            word.append(char)
    _word = ''.join(word)
    parts.append(_word)

    return '_'.join(parts[1:])


def snake_to_camel(string):
    parts = string.split('_')
    camelized = ''.join(f'{x[0].upper()}{x[1:]}' for x in parts)
    return f'{parts[0]}{camelized}'



from ._base import BaseMixin, Endpoint
from ._mixins import DisableNotification, ReplyMarkup, ReplyToMessageId


class VenueMixin(BaseMixin):
    def foursquare_id(self, foursquare_id):
        self.args['foursquare_id'] = foursquare_id
        return self

    def foursquare_type(self, foursquare_type):
        self.args['foursquare_type'] = foursquare_type
        return self


class SendVenue(Endpoint, DisableNotification, ReplyMarkup, ReplyToMessageId, VenueMixin):
    def __init__(self, chat_id, latitude, longitude, title, address):
        super().__init__()
        self.args['chat_id'] = chat_id
        self.args['latitude'] = latitude
        self.args['longitude'] = longitude
        self.args['title'] = title
        self.args['address'] = address



from ._base import InlineQueryResult, InlineQueryResultCachedFile, InlineQueryResultFile
from ._mixins import Description, Duration, InputMessageContent, Size, Thumb, Title
from .._mixins import Caption, ParseMode
from ..contact import ContactMixin
from ..location import LocationMixin
from ..venue import VenueMixin


class Article(InlineQueryResult, Description, ParseMode, Thumb):
    def __init__(self, title, input_message_content):
        super().__init__()
        self.args['title'] = title
        self.args['input_message_content'] = input_message_content

    def url(self, url):
        self.args['url'] = url
        return self

    def hide_url(self):
        self.args['hide_url'] = True
        return self


#
class _Photo(Caption, Description, InputMessageContent, ParseMode, Title):
    pass


class Photo(InlineQueryResultFile, _Photo, Size):
    pass


class CachedPhoto(InlineQueryResultCachedFile, _Photo):
    pass


#
class _Gif(Caption, InputMessageContent, ParseMode, Title):
    pass


class Gif(InlineQueryResultFile, _Gif, Duration, Size):
    pass


class CachedGif(InlineQueryResultCachedFile, _Gif):
    pass


#
class _Mpeg4Gif(Caption, InputMessageContent, ParseMode, Title):
    pass


class Mpeg4Gif(InlineQueryResultFile, _Mpeg4Gif, Duration, Size):
    def __init__(self, url, thumb_url):
        super().__init__(url, thumb_url)
        self.prefix = 'mpeg4'


class CachedMpeg4Gif(InlineQueryResultCachedFile, _Mpeg4Gif):
    def __init__(self, file_id):
        super().__init__(file_id)
        self.prefix = 'mpeg4'


#
class _Video(Caption, Description, InputMessageContent, ParseMode):
    pass


class Video(InlineQueryResult, _Video, Size):
    mime_types = {'text/html', 'video/mp4'}

    def __init__(self, url, mime_type, thumb_url, title):
        assert mime_type in Video.mime_types

        super().__init__()
        self.args['video_url'] = url
        self.args['mime_type'] = mime_type
        self.args['thumb_url'] = thumb_url
        self.args['title'] = title


class CachedVideo(InlineQueryResult, _Video):
    def __init__(self, file_id, title):
        super().__init__()
        self.args['video_file_id'] = file_id
        self.args['title'] = title


#
class _Audio(Caption, InputMessageContent, ParseMode):
    pass


class Audio(InlineQueryResultFile, _Audio, Duration):
    def performer(self, performer):
        self.args['performer'] = performer
        return self


class CachedAudio(InlineQueryResultCachedFile):
    pass


#
class _Voice(Caption, InputMessageContent, ParseMode):
    pass


class Voice(InlineQueryResult, _Voice):
    def __init__(self, voice_url, title):
        super().__init__()
        self.args['voice_url'] = voice_url
        self.args['title'] = title

    def voice_duration(self, voice_duration):
        self.args['voice_duration'] = voice_duration
        return self


class CachedVoice(InlineQueryResult, _Voice):
    def __init__(self, file_id, title):
        super().__init__()
        self.args['voice_file_id'] = file_id
        self.args['title'] = title


#
class _Document(Caption, Description, InputMessageContent, ParseMode):
    pass


class Document(InlineQueryResult, _Document, Thumb):
    mime_types = {'application/pdf', 'application/zip'}

    def __init__(self, title, document_url, mime_type):
        assert mime_type in Document.mime_types

        super().__init__()
        self.args['title'] = title
        self.args['document_url'] = document_url
        self.args['mime_type'] = mime_type


class CachedDocument(InlineQueryResult, _Document):
    def __init__(self, title, file_id):
        super().__init__()
        self.args['title'] = title
        self.args['document_file_id'] = file_id


#
class Location(InlineQueryResult, InputMessageContent, LocationMixin, Thumb):
    def __init__(self, latitude, longitude, title):
        super().__init__()
        self.args['latitude'] = latitude
        self.args['longitude'] = longitude
        self.args['title'] = title


class Venue(InlineQueryResult, InputMessageContent, VenueMixin, Thumb):
    def __init__(self, latitude, longitude, title, address):
        super().__init__()
        self.args['latitude'] = latitude
        self.args['longitude'] = longitude
        self.args['title'] = title
        self.args['address'] = address


class Contact(InlineQueryResult, ContactMixin, InputMessageContent, Thumb):
    def __init__(self, phone_number, first_name):
        super().__init__()
        self.args['phone_number'] = phone_number
        self.args['first_name'] = first_name


class Game(InlineQueryResult):
    def __init__(self, game_short_name):
        super().__init__()
        self.args['game_short_name'] = game_short_name


class CachedSticker(InlineQueryResult, InputMessageContent):
    def __init__(self, file_id):
        super().__init__()
        self.args['sticker_file_id'] = file_id



from abc import ABC, ABCMeta

from .._mixins import ReplyMarkup
from ..._utils.casing import camel_to_snake
from ..._utils.random import random_string


class InlineQueryResultMixin(ABC):
    __slots__ = []
    _type: str
    prefix: str
    args: dict


class InlineQueryResult(ReplyMarkup, metaclass=ABCMeta):
    __slots__ = ['_type', 'prefix', 'args']

    def __init__(self):
        classname = self.__class__.__name__
        self._type = camel_to_snake(classname)
        self.prefix = self._type

        self.args = {
            'type': self._type,
            'id': random_string(8)
        }


class InlineQueryResultFile(InlineQueryResult, metaclass=ABCMeta):
    def __init__(self, url, thumb_url):
        super().__init__()
        url_key = f'{self.prefix}_url'
        self.args[url_key] = url
        self.args['thumb_url'] = thumb_url


class InlineQueryResultCachedFile(InlineQueryResult, metaclass=ABCMeta):
    def __init_(self, file_id):
        super().__init__()
        self._type = self._type[8:]  # "cached_"
        file_id_key = f'{self.prefix}_file_id'
        self.args[file_id_key] = file_id


class InputMessageContent(ABC):
    __slots__ = ['args']



from ._base import InlineQueryResultMixin
from ...settings import json_lib as json


class Description(InlineQueryResultMixin):
    def description(self, description):
        self.args['description'] = description
        return self


class Duration(InlineQueryResultMixin):
    def duration(self, duration):
        key = f'{self.prefix}_duration'
        self.args[key] = duration
        return self


class InputMessageContent(InlineQueryResultMixin):
    def input_message_content(self, input_message_content):
        self.args['input_message_content'] = json.dumps(input_message_content.args)
        return self


class Size(InlineQueryResultMixin):
    def width(self, width):
        key = f'{self.prefix}_width'
        self.args[key] = width
        return self

    def height(self, height):
        key = f'{self.prefix}_height'
        self.args[key] = height
        return self


class Title(InlineQueryResultMixin):
    def title(self, title):
        self.args['title'] = title
        return self


class Thumb(InlineQueryResultMixin):
    def thumb_url(self, thumb_url):
        self.args['thumb_url'] = thumb_url
        return self

    def thumb_width(self, thumb_width):
        self.args['thumb_width'] = thumb_width
        return self

    def thumb_height(self, thumb_height):
        self.args['thumb_height'] = thumb_height
        return self



from ._base import File
from ._mixins import Thumb, ToInputMedia
from .._mixins import Caption, ParseMode


class Photo(File, Caption, ParseMode, ToInputMedia):
    @staticmethod
    def _get_file_id(response):
        return response[-1]


class Audio(File, Caption, ParseMode, Thumb, ToInputMedia):
    def duration(self, duration):
        self.args['duration'] = duration
        return self

    def performer(self, performer):
        self.args['performer'] = performer
        return self

    def title(self, title):
        self.args['title'] = title
        return self


class Document(File, Caption, ParseMode, Thumb, ToInputMedia):
    pass


class Video(File, Caption, ParseMode, Thumb, ToInputMedia):
    def supports_streaming(self):
        self.args['supports_streaming'] = True
        return self

    def duration(self, duration):
        self.args['duration'] = duration
        return self

    def width(self, width):
        self.args['width'] = width
        return self

    def height(self, height):
        self.args['height'] = height
        return self


class Animation(File, Caption, ParseMode, Thumb, ToInputMedia):
    def duration(self, duration):
        self.args['duration'] = duration
        return self

    def width(self, width):
        self.args['width'] = width
        return self

    def height(self, height):
        self.args['height'] = height
        return self


class Voice(File, Caption, ParseMode):
    def duration(self, duration):
        self.args['duration'] = duration
        return self


class VideoNote(File, Thumb):
    def duration(self, duration):
        self.args['duration'] = duration
        return self

    def length(self, length):
        self.args['length'] = length
        return self



import asyncio

from .http11 import build_request


class TCP(asyncio.Protocol):
    def connection_made(self, transport):
        self.transport = transport

    def connection_lost(self, exc):
        # the transport is already close when we get here
        if exc:
            self.future.set_exception(exc)

    def data_received(self, data):
        # since:
        # 1) http parsing in python is _extremely_ inefficient compared to more modern solutions
        # 2) making the assumption that all the data will fit in a single socket.read and there will not be chunks both
        # speeds up and simplifies this code a lot
        # 3) we can safely make that assumption in the context of the Telegram bot API
        # this code doesn't (shouln't) need to be longer than this:
        self.future.set_result(data)

    def eof_received(self):
        # falsy -> the transport closes itself
        # truthy -> the protocol determines whether to close the transport
        # SSL doesn't support half-closed streams so here we go
        return False

    def write(self, data, future):
        self.future = future
        self.transport.write(data)


class HttpConnectionPool:
    def __init__(self, url, *, loop=None, size=5):
        self.url = url

        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        self.queue = asyncio.Queue(loop=loop)

        for _ in range(size):
            loop.create_task(self.open())

    async def open(self):
        transport = None

        while True:
            (future, req) = await self.queue.get()
            if transport is None:
                (transport, protocol) = await self.loop.create_connection(lambda: TCP(), self.url, 443, ssl=True)

            protocol.write(req, future)
            await future
            if self.queue.empty():
                transport.close()
                transport = None

    def request(self, path, method, headers, body):
        req = build_request(self.url, path, method, headers, body)
        future = self.loop.create_future()
        self.queue.put_nowait((future, req))
        return future



from uuid import uuid4

from ._base import FileMixin
from ..._utils.multipart_encoder import MultipartEncoder


class Thumb(FileMixin):
    def thumb(self, thumb):

        # thumb must be a path
        if self.is_data:
            self.value.append(MultipartEncoder.encode_file('thumb', thumb))

        return self


class ToInputMedia(FileMixin):
    media_types = {'photo', 'video'}

    def to_input_media(self):
        serialized = {
            'type': self.file_type,
            **self.args,
            **self.attributes
        }

        if self.is_data:
            files = []
            for file in self.value:

                # avoid conflicts
                field_name = uuid4().hex

                # if it's neither a photo nor a video it's a thumb
                if file[0] in self.media_types:
                    serialized['media'] = f'attach://{field_name}'
                else:
                    serialized[file[0]] = f'attach://{field_name}'

                files.append((field_name, file[1], file[2], file[3]))
            return serialized, files
        else:
            serialized['media'] = self.value
            return serialized, None



from ._base import Endpoint


class GetMe(Endpoint):
    pass


class GetUserProfilePhotos(Endpoint):
    def __init__(self, user_id):
        super().__init__()
        self.args['user_id'] = user_id

    def offset(self, offset):
        self.args['offset'] = offset
        return self

    def limit(self, limit):
        self.args['limit'] = limit
        return self
